###
### staticcharme.py
###
### Charme with Static Typing
### David Evans

###
### Parser
###

def parseError(msg):
    print "Parse error: " + msg
    
def tokenize(s):
    current = ''
    tokens = []
    for c in s:
        if c.isspace():
            if len(current) > 0:
                tokens.append(current)
                current = ''
        elif c in '():': # added : to support type declarations
            if len(current) > 0:
                tokens.append(current)
                current = ''
            tokens.append(c)
        else:
            current = current + c

    if len(current) > 0:
        tokens.append(current)
    return tokens

def parse(s):
    def parsetokens(tokens, inner):
       res = []
       while len(tokens) > 0:
          current = tokens.pop(0)
          if current == '(':
             res.append (parsetokens(tokens, True))
          elif current == ')':
             if inner:
                return res
             else:
                parseError("Unmatched close paren: " + s)
                return None
          else:
             res.append(current)
        
       if inner:
          parseError ("Unmatched open paren: " + s)
          return None
       else:
          return res

    return parsetokens(tokenize(s), False)

###
### Evaluator
###

def evalError(msg):
    print "Error: " + msg    
    
###
### Primitive Procedures
###
    
def checkOperands(operands, num, prim):
    if (len(operands) != num):
       print ("%s expected %s operands, given %s: %s"
              % (prim, num, len(operands), str(operands)))
           
def primitivePlus (operands):
    if (len(operands) == 0):
       return 0
    else:
       return operands[0] + primitivePlus (operands[1:])

def primitiveTimes (operands):
    if (len(operands) == 0):
       return 1
    else:
       return operands[0] * primitiveTimes (operands[1:])
    
def primitiveMinus (operands):
    if (len(operands) == 1):
       return -1 * operands[0]
    elif len(operands) == 2:
       return operands[0] - operands[1]
    else:
       evalError("Primitive - expected 1 or 2 operands, given %s: %s"
                  % (len(operands), str(operands)))

def primitiveEquals (operands):
    checkOperands (operands, 2, "=")
    return operands[0] == operands[1]

def primitiveZero (operands):
    checkOperands (operands, 1, "zero?")
    return operands[0] == 0

def primitiveGreater (operands):
    checkOperands (operands, 2, ">")
    return operands[0] > operands[1]

def primitiveLessThan (operands):
    checkOperands (operands, 2, "<")
    return operands[0] < operands[1]

###
### Primitives
###

def isNumber(expr):
    return isinstance(expr, str) and expr.isdigit()

def isPrimitiveProcedure(expr):
    return callable(expr)

def isPrimitive(expr):
    return (isNumber(expr) or isPrimitiveProcedure(expr))

def evalPrimitive(expr):
    # A primitive evaluates to its pre-defined value
    if isNumber(expr):
        return int(expr)
    else:
        return expr        

def typePrimitive(expr):
    if isNumber(expr):
        return CPrimitiveType('Number')
    elif isinstance(expr, bool):
    	return CPrimitiveType('Boolean')
    elif callable(expr):
        return globalEnvironment.reverseLookupType(expr)
    else:
        assert False
        
###
### Conditionals
###
    
def isSpecialForm(expr, keyword):
    return isinstance(expr, list) \
           and len(expr) > 0 and expr[0] == keyword

def isConditional(expr):
    return isSpecialForm(expr, 'cond')

def evalConditional(expr, env):
    assert isConditional(expr)
    if len(expr) <= 2:
        evalError ("Bad conditional expression: %s" % str(expr))
    for clause in expr[1:]:
        if len(clause) != 2:
            evalError ("Bad conditional clause: %s" % str(clause))
        predicate = clause[0]
        result = meval(predicate, env)
        if result:
            return meval(clause[1], env)
    evalError ("No conditional predicate evaluates to non-false value: %s"
               % (expr))
    return None

def typeConditional(expr, env):
    assert isConditional(expr)
    if len(expr) <= 2:
        evalError ("Bad conditional expression: %s" % str(expr))
    firstclause = True
    for clause in expr[1:]:
        if len(clause) != 2:
            evalError ("Bad conditional clause: %s" % str(clause))
        predicate = clause[0]
        if not typecheck(predicate, env).matches(CPrimitiveType('Boolean')):
            return CErrorType ("Non-Boolean predicate: " + str(predicate))
        consequent = clause[1]
        if firstclause:
            clausetype = typecheck(consequent, env)
            firstclause = False
        else:
            if not typecheck(consequent, env).matches(clausetype):
                return CErrorType ("Mistyped consequent: " + str(consequent))
    return clausetype

###
### Begins
###

def isBegin(expr):
    return isSpecialForm(expr, 'begin')

def evalBegin(expr, env):
    # Question 8

def typeBegin(expr, env):
    # Stub implementation (replace this for Question 9)
    return CPrimitiveType('Number')

###
### Names and Definitions
###
    
class Environment:
    # To support typed names, we store a [type, value] pair for each variable.
    def __init__(self, parent):
        self._parent = parent
        self._frame = { }
    def addVariable(self, name, typ, value):
        # add a new name, or replace old value
        # store a tuple of the type and value
        self._frame[name] = (typ, value)
    def _lookupPlace(self, name):
        if self._frame.has_key(name):
            return self._frame[name]
        elif (self._parent):
            return self._parent._lookupPlace(name)
        else:
            return None
    def lookupVariable(self, name):
        return self._lookupPlace(name)[1]
    def lookupVariableType(self, name):
        place = self._lookupPlace(name)
        if place:
            return place[0]
        else:
            return CErrorType("Name not found: " + name)
    def reverseLookupType(self, val):
        # This is a bit of a kludge for dealing with primitive
        # procedure types.
        for entry in self._frame.itervalues():
            if entry[1] == val:
                return entry[0]
        return CErrorType("Value not found")            
    def __str__(self):
        return "<Environment: %s / %s>" % (str(self._frame), str(self._parent))

def isDefinition(expr):
    return isSpecialForm(expr, 'define')

def evalDefinition(expr, env):
    assert isDefinition(expr)
    name = expr[1]
    if isinstance(name, str):
         value = meval(expr[4], env)
         typ = parseType(expr[3])
         env.addVariable(name, typ, value)
    elif isinstance(name, list):
        evalError ("Procedure definition syntax not implemented")
    else:
        evalError ("Bad definition: %s" % str(expr))

def typeDefinition(expr, env):
    assert isDefinition(expr)
    if len(expr) != 5:
        evalError ("Bad definition: %s" % str(expr))
    name = expr[1]
    if isinstance(name, str):
        if expr[2] != ':':
            return CErrorType ("Definition missing type: %s" % str(expr))        
        typ = parseType(expr[3])
        newenv = Environment(env)
        newenv.addVariable(name,typ,None)
        etyp = typecheck(expr[4], newenv)
        if not typ.matches(etyp):
            return CErrorType ("Mistyped definition: %s declared type %s, actual type %s"
                                    % (name, typ, etyp))
        return None # definition has no type
    elif isinstance(name, list):
        return CErrorType ("Procedure definition syntax not implemented")
    else:
        return CErrorType ("Bad definition: %s" % str(expr))
    
def isName(expr):
    return isinstance(expr, str)

def evalName(expr, env):
    # To evaluate a name, lookup the value of the name in the environment
    assert isName(expr)
    return env.lookupVariable(expr)

def typeName(expr, env):
    return env.lookupVariableType(expr)

###
### Procedures and Applications
###

class Procedure:
    def __init__(self, params, typ, body, env):
        self._params = params
        self._body = body
        self._typ = typ
        self._env = env
    def getParams(self):
        return self._params
    def getParamTypes(self):
        return self._typ
    def getBody(self):
        return self._body
    def getEnvironment(self):
        return self._env        
    def __str__(self):
        return "<Procedure %s / %s>" % (str(self._params), str(self._body))

def isLambda(expr):
    return isSpecialForm(expr, 'lambda')

def evalLambda(expr, env):
    assert isLambda(expr)
    if len(expr) != 3:
        evalError ("Bad lambda expression: %s" % (str(expr)))
    params = expr[1]
    paramtypes = []
    paramnames = []
    assert len(params) % 3 == 0
    for i in range(0, len(params) / 3):
        name = params[i*3]
        assert params[(i*3)+1] == ':'
        paramnames.append(name)
        typ = parseType(params[(i*3)+2])
        paramtypes.append(typ)
    return Procedure(paramnames, paramtypes, expr[2], env)

def typeLambda(expr, env):
    assert isLambda(expr)
    if len(expr) != 3:
        evalError ("Bad lambda expression: %s" % str(expr))
    # this is a bit tricky - we need to "partially" apply it
    # to find the type of the body
    newenv = Environment(env)
    params = expr[1]
    paramnames = []
    paramtypes = []
    assert len(params) % 3 == 0
    for i in range(0, len(params) / 3):
        name = params[i*3]
        assert params[(i*3)+1] == ':'
        typ = parseType(params[(i*3)+2])
        paramnames.append(name)
        paramtypes.append(typ)
        # No values, but need to initialize types
        newenv.addVariable(name, typ, None)
    resulttype = typecheck(expr[2], newenv)
    return CProcedureType(CProductType(paramtypes), resulttype)
    
def isApplication(expr):
    # Applications are lists [proc, oper, oper]
    # Assumes expr is not a special form (must check special forms first)
    return isinstance(expr, list)
   
def evalApplication(expr, env):
    # To evaluate an application, evaluate all the subexpressions
    subexprs = expr
    subexprvals = map (lambda sexpr: meval(sexpr, env), subexprs)
    # then, apply the value of the first subexpression to the rest
    return mapply(subexprvals[0], subexprvals[1:])

def typeApplication(expr, env):
    proctype = typecheck(expr[0], env)
    if not proctype.isProcedureType():
        return CErrorType("Application of non-procedure: " + str(expr[0]))
    optypes = map (lambda op: typecheck(op, env), expr[1:])
    optype = CProductType(optypes)
    if not proctype.getParameterTypes().matches(optype):
        return CErrorType("Parameter type mismatch: expected %s, given %s" \
                          % (proctype.getParameterTypes(), optype))
    return proctype.getReturnType()

def mapply(proc, operands):
    if (isPrimitiveProcedure(proc)):
        return proc(operands)
    elif isinstance(proc, Procedure):
        params = proc.getParams()
        types = proc.getParamTypes()
        newenv = Environment(proc.getEnvironment())
        if len(params) != len(operands):
             evalError ("Parameter length mismatch: %s given operands %s" \
                        % (str(proc), str(operands)))
        for i in range(0, len(params)):
             newenv.addVariable(params[i], types[i], operands[i])        
        return meval(proc.getBody(), newenv)        
    else:
        evalError("Application of non-procedure: %s" % (proc))

###
### Core Evaluator
###

def meval(expr, env):
    if isPrimitive(expr):
       return evalPrimitive(expr)
    elif isConditional(expr):
       return evalConditional(expr, env)
    elif isLambda(expr):
       return evalLambda(expr, env)
    elif isDefinition(expr):                
       evalDefinition(expr, env)
    elif isBegin(expr):
       return evalBegin(expr, env)
    elif isName(expr):
       return evalName(expr, env)
    elif isApplication(expr):
       return evalApplication(expr, env)
    else:
       evalError ("Unknown expression type: " + str(expr))

def typecheck(expr, env):
    if isPrimitive(expr):
        return typePrimitive(expr)
    elif isConditional(expr):
        return typeConditional(expr, env)
    elif isLambda(expr):
        return typeLambda(expr, env)
    elif isDefinition(expr):
        return typeDefinition(expr, env)
    elif isBegin(expr):
        return typeBegin(expr, env)
    elif isName(expr):
        return typeName(expr, env)
    elif isApplication(expr):
        return typeApplication(expr, env)
    else:
        evalError ("Unknown expression type: " + str(expr))

def parseType(p):
    if isinstance(p, str):
        if p == 'Number': return CPrimitiveType('Number')
        elif p == 'Boolean': return CPrimitiveType('Boolean')
        else: evalError("Undefined type: " + p)
    else:
        if len(p) == 3 and p[1] == '->':
            return CProcedureType(parseType(p[0]), parseType(p[2]))
        else:
            types = map(parseType, p)
            return CProductType(types)

def typeFromString(s):
    p = parse(s)
    assert len(p) == 1
    return parseType(p[0])

class CType:
    # These methods are overridden by subclasses
    def isPrimitiveType(self):
        return False
    def isProcedureType(self):
        return False
    def isProductType(self):
        return False
    def isError(self):
        return False
    def matches(self, other):
        # subclass method needs to match ("abstract" method)
        assert False

class CPrimitiveType(CType):
    def __init__(self, s):
        self._name = s
    def __str__(self):
        return self._name
    def isPrimitiveType(self):
        return True
    def matches(self, other):
        return other.isPrimitiveType() and self._name == other._name

class CProcedureType(CType):
    def __init__(self, args, rettype):
        self._args = args
        self._rettype = rettype
    def __str__(self):
        return "(" + str(self._args) + " -> " + str(self._rettype) + ")"
    def isProcedureType(self):
        return True
    def getReturnType(self):
        return self._rettype
    def getParameterTypes(self):
        return self._args
    def matches(self, other):
        return other.isProcedureType() \
               and self._args.matches(other._args) \
               and self._rettype.matches(other._rettype)

class CProductType(CType):
    def __init__(self, types):
        self._types = types
    def __str__(self):
        res = "("
        firstone = True
        for t in self._types:
            if firstone:
                firstone = False
            else:
                res = res + " "
            res = res + str(t)
        res = res + ")"
        return res
    def isProductType(self):
        return True
    def matches(self, other):
        if other.isProductType():
            st = self._types
            ot = other._types
            if len(st) == len(ot):
                for i in range(0, len(st)):
                    if not st[i].matches(ot[i]):
                        return False
                # reached end of loop ==> all matched
                return True
            else:
                return False
        else:
            return False

class CErrorType(CType):
    def __init__(self, msg):
        self._msg = msg
    def __str__(self):
        return "<Type Error: " + self._msg + ">"
    def getMessage(self):
        return self._msg
    def matches(self, other):
        return False
    def isError(self):
        return True

def initializeGlobalEnvironment():
    global globalEnvironment                  
    globalEnvironment = Environment(None)
    globalEnvironment.addVariable('#t', CPrimitiveType('Boolean'), True)
    globalEnvironment.addVariable('#f', CPrimitiveType('Boolean'), False)                   
    globalEnvironment.addVariable('+', typeFromString('((Number Number) -> Number)'), primitivePlus)
    globalEnvironment.addVariable('-', typeFromString('((Number Number) -> Number)'), primitiveMinus)
    globalEnvironment.addVariable('*', typeFromString('((Number Number) -> Number)'), primitiveTimes)
    globalEnvironment.addVariable('=', typeFromString('((Number Number) -> Boolean)'), primitiveEquals)
    globalEnvironment.addVariable('zero?', typeFromString('((Number) -> Boolean)'), primitiveZero)
    globalEnvironment.addVariable('>', typeFromString('((Number Number) -> Boolean)'), primitiveGreater)
    globalEnvironment.addVariable('<', typeFromString('((Number Number) -> Boolean)'), primitiveLessThan)

def evalLoop():
    print "Welcome to StaticCharme (the Cautious Snake Charmer's Language)"
    initializeGlobalEnvironment()
    while True:
        intext = raw_input("StaticCharme> ")
        commands = intext.split()
        if commands[0] == 'quit':
            break
        elif commands[0] == 'load':
            if len(commands) != 2:
                print "Load commands expects one parameter."
                continue
            else:
                fname = commands[1]
                try:
                    f = open(fname)
                    intext = f.read()
                    f.close()                
                except:
                    print "Error reading file: " + fname
        elif commands[0] == 'type':
            exprs = parse(intext)
            if exprs:
                for expr in exprs[1:]:
                    # first, typecheck the expression
                    typ = typecheck(expr, globalEnvironment)
                    if typ and typ.isError():
                        print "Error: " + typ.getMessage()
                    else:
                        print str(typ)
            continue
        try:
            exprs = parse(intext)
            if exprs:
                for expr in exprs:
                    # first, typecheck the expression
                    typ = typecheck(expr, globalEnvironment)
                    if typ and typ.isError():
                        print "Error: " + typ.getMessage()
                    else:
                        res = meval(expr, globalEnvironment)
                        if res != None:
                            print str(res)
        except:
            print "Error evaluating " + str(intext)
            raise
